;;;
;;; circuit-components.lisp
;;;
;;; Basic circuit elements for circuit-sim.lisp
;;;
;;; $Revision: 1.2 $
;;;
;;; $Log: circuit-components.lisp,v $
;;; Revision 1.2  2007/02/11 18:46:44  jao
;;; 1. In package "CIRCUIT"
;;; 2. Use definitions from circuit-logic.lisp for operations.
;;;
;;; Revision 1.1  2007/02/11 00:39:20  jao
;;; Initial revision
;;;
;;;

(in-package "CIRCUIT")

(defun inverter (input output propagation-delay)
  (labels ((invert-input ()
             ;; (format t "Invert input: ~D " (signal-value input))
             (let ((new-value (logical-not (get-signal input))))
               ;; (format t "new output value: ~D~%" new-value)
               (after-delay propagation-delay 
			    (set-signal output new-value)))))
    (add-action input #'invert-input)))

(defun nand-gate (in1 in2 output propagation-delay)
  (labels ((nand-inputs ()
             (let ((new-value (logical-nand 
			       (get-signal in1) 
			       (get-signal in2))))
               (after-delay propagation-delay
                            (set-signal output new-value)))))
    (add-action in1 #'nand-inputs)
    (add-action in2 #'nand-inputs)))
	   
#||
;;; test nand gate
;;; see circuit-tests::test-nand-gate
 
(progn
  (setf *agenda* (make-agenda 0)
	*d1* (make-wire)
	*d2* (make-wire)
	*out* (make-wire))
  (nand-gate *d1* *d2* *out* 10)
  (probe *out* "nand output")
  
  (at-time 0 (set-signal *d1* 0)
	   (set-signal *d2* 0)
	   (format t "~%~%NAND(0,0)~%"))
  
  (at-time 30 (set-signal *d1* 1)
	   (format t "NAND(1,0)~%"))

  (at-time 50 (set-signal *d2* 1)
	   (format t "NAND(1,1)~%"))
  (at-time 70 (set-signal *d1* 0)
	   (format t "NAND(0,1)~%"))

  (propagate))
||#    

(defun and-gate (in1 in2 output propagation-delay)
  (labels ((and-inputs ()
             (let ((new-value (logical-and (get-signal in1) (get-signal in2))))
               (after-delay propagation-delay
                            (set-signal output new-value)))))
    (add-action in1 #'and-inputs)
    (add-action in2 #'and-inputs)))

(defun or-gate (in1 in2 output propagation-delay)
  (labels ((or-inputs ()
             (let ((new-value (logical-or (get-signal in1) (get-signal in2))))
               (after-delay propagation-delay
                            (set-signal output new-value)))))
    (add-action in1 #'or-inputs)
    (add-action in2 #'or-inputs)))

(defun half-adder (in1 in2 s c or-prop-delay and-prop-delay inv-prop-delay)
  (let ((d (make-wire))
        (e (make-wire)))
    (or-gate in1 in2 d or-prop-delay)
    (and-gate in1 in2 c and-prop-delay)
    (inverter c e inv-prop-delay)
    (and-gate d e s and-prop-delay)))

(defun full-adder (in1 in2 carry-in sum carry-out or-prop-delay and-prop-delay inv-prop-delay)
  (let ((s (make-wire))
        (c1 (make-wire))
        (c2 (make-wire)))
    (half-adder in2 carry-in s c1 or-prop-delay and-prop-delay inv-prop-delay)
    (half-adder in1 s sum c2 or-prop-delay and-prop-delay inv-prop-delay)
    (or-gate c1 c2 carry-out or-prop-delay)))

(defun d-flipflop (data-in clock-in d-output clock-propagation-delay 
                           &key (initial-d-output 0) (clock-edge :rising) 
			   (data-in-setup-time 0) (data-in-hold-time 0)
                           (data-propagation-delay 0))
  "Clock-edge should be :rising or :falling. data-propagation-delay only matters for non-zero hold time."

  ;;;
  ;;; HOW TO IMPLEMENT SETUP/HOLD TIME REQUIREMENT?? -- catch changes on 
  ;;; data-in, record them. (Seems to work.)
  ;;;
  (let ((last-data-in-change (agenda-time *agenda*))
        (last-clock-edge (agenda-time *agenda*)))

    (labels ((data-process () ;; Process events on the D input
	       (let* ((current-time (agenda-time *agenda*))
		      (data-in-held-time (- current-time last-clock-edge)))
		 (setf last-data-in-change current-time)
		 (when (< data-in-held-time data-in-hold-time)
                   (warn "Missed hold-time requirement")
                   (after-delay data-propagation-delay 
                            (set-signal 
			     d-output 
			     :invalid-hold-time)))))

             (clock-process () ;; Process events on the Clock input
               (let ((current-clock (get-signal clock-in))
                     (data-in-was-setup-time (- (agenda-time *agenda*) 
						last-data-in-change)))
		 
                 ;; know we got triggered on a clock change. Let's
		 ;; determine if it is the correct edge 
		 
                 (when (or (and (eq clock-edge :rising) 
				(or (eq current-clock t) 
				    (= current-clock 1))) 
                           (and (eq clock-edge :falling) 
				(or (eq current-clock nil) 
				    (= current-clock 0))))
		   
                   (setf last-clock-edge (agenda-time *agenda*))
                   (if (< data-in-was-setup-time data-in-setup-time)
		       
		       (progn (warn "Missed setup-time requirement")
			      (after-delay 
			       clock-propagation-delay
			       (set-signal
				d-output
				:invalid-setup-time)))

		       (after-delay clock-propagation-delay
				    (set-signal 
				     d-output
				     (get-signal data-in))))))))

  ;;; add-action is not really the right way to initialize the
  ;;; output---it might detect a clock-edge when the flip-flop is 
  ;;; added, or it might not.
  ;;; 
  ;;; ought to specify the initial state of the d-output
  ;;;

      (add-action clock-in #'clock-process (list #'(lambda ()))) 
      (add-action data-in #'data-process (list #'(lambda ()
						   (set-signal 
						    d-output
						    initial-d-output)))))))

(defun d-flipflop-jamclear (data-in clock-in d-output clear
			    clock-propagation-delay clear-propagation-delay
			    &key (initial-d-output 0) (clock-edge :rising) 
			    (data-in-setup-time 0) (data-in-hold-time 0)
			    (data-propagation-delay 0))
  "Clock-edge should be :rising or :falling. data-propagation-delay only matters for non-zero hold time."

  ;;;
  ;;; HOW TO IMPLEMENT SETUP/HOLD TIME REQUIREMENT?? -- catch changes on 
  ;;; data-in, record them. (Seems to work.)
  ;;;
  (let ((last-data-in-change (agenda-time *agenda*))
        (last-clock-edge (agenda-time *agenda*)))

    (labels ((data-process () ;; Process events on the D input
	       (unless (logical-false (get-signal clear))
		 (let* ((current-time (agenda-time *agenda*))
			(data-in-held-time (- current-time last-clock-edge)))
		   (setf last-data-in-change current-time)
		   (when (< data-in-held-time data-in-hold-time)
		     (warn "Missed hold-time requirement")
		     (after-delay data-propagation-delay 
				  (set-signal 
				   d-output 
				   :invalid-hold-time))))))

             (clock-process () ;; Process events on the Clock input
               (unless (logical-false (get-signal clear))
		 (let ((current-clock (get-signal clock-in))
		       (data-in-was-setup-time (- (agenda-time *agenda*) 
						  last-data-in-change)))
		   
		   ;; know we got triggered on a clock change. Let's
		   ;; determine if it is the correct edge 
		   
		   (when (or (and (eq clock-edge :rising) 
				  (or (eq current-clock t) 
				      (= current-clock 1))) 
			     (and (eq clock-edge :falling) 
				  (or (eq current-clock nil) 
				      (= current-clock 0))))
		     
		     (setf last-clock-edge (agenda-time *agenda*))
		     (if (< data-in-was-setup-time data-in-setup-time)
			 
			 (progn (warn "Missed setup-time requirement")
				(after-delay 
				 clock-propagation-delay
				 (set-signal
				  d-output
				  :invalid-setup-time)))
			 
			 (after-delay clock-propagation-delay
				      (set-signal 
				       d-output
				       (get-signal data-in))))))))

	     (clear-process ()
	       (if (logical-false (get-signal clear))
		   (after-delay clear-propagation-delay
				(set-signal d-output
					    0)))))
	     
  ;;; add-action is not really the right way to initialize the
  ;;; output---it might detect a clock-edge when the flip-flop is 
  ;;; added, or it might not.
  ;;; 
  ;;; ought to specify the initial state of the d-output
  ;;;


      (add-action clock-in #'clock-process (list #'(lambda ()))) 
      (add-action data-in #'data-process (list #'(lambda ()
						   (set-signal 
						    d-output
						    initial-d-output))))
      (add-action clear #'clear-process)))) 


(defun ganged-d-flipflops (data-in-bus clock-in output-bus clear
			   clock-propagation-delay 
			   clear-propagation-delay
			   &key (initial-d-output 0) (clock-edge :rising) 
			   (data-in-setup-time 0) (data-in-hold-time 0)
                           (data-propagation-delay 0))
  "Multiple flip-flops, sharing a clock and clear"
  
  (let ((in-width (bus-width data-in-bus))
	(out-width (bus-width output-bus)))
    (unless (= in-width out-width)
      (error "Ganged flip-flops must have the same number of inputs and outputs."))
    (dotimes (i in-width)
      (d-flipflop-jamclear (nth-wire i data-in-bus)
			   clock-in
			   (nth-wire i output-bus)
			   clear
			   clock-propagation-delay 
			   clear-propagation-delay
			   :initial-d-output initial-d-output
			   :clock-edge clock-edge
			   :data-in-setup-time data-in-setup-time
			   :data-in-hold-time data-in-hold-time
			   :data-propagation-delay data-propagation-delay))))